#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

#define BAUDRATE B115200

#define SENDSIZE 1024

int ttyfd;

int SendCMD(const char *fmt, ...);

int bSend (unsigned char data, unsigned int address, int isflush);

int AwaitACK(void);
int ReadMOT(FILE *fp, int (*bwrite)(unsigned char data, unsigned int address, int isflush));

int fdputc(int tty, unsigned char c) 
{
  return write(tty, &c, 1);
}
int fdput16(int tty, unsigned short i)
{
  fdputc(tty, i >> 8); //MSB first
  fdputc(tty, i % 0x100);
  return 0;
}
int fdput24(int tty, unsigned int i)
{
  fdput16(tty, i >> 8); //MSB first
  fdputc(tty, i % 0x100);
  return 0;
}

unsigned short MakeCheckSum(unsigned char *buf, int cnt);

int main(int argc, char **argv) 
{
  FILE *srcfp;
  struct termios ttyios;
  int ret;
  char *s;
  if((ttyfd = open(argv[3], O_RDWR)) == -1)
    {
      fprintf(stderr, "tty open failed\n");
      return -1;
    }
  
  tcsetpgrp(ttyfd, getpgrp());
  tcgetattr(ttyfd, &ttyios);
  cfsetispeed(&ttyios, BAUDRATE);
  cfsetospeed(&ttyios, BAUDRATE);
  cfmakeraw(&ttyios);
  ttyios.c_cflag |= HUPCL | CLOCAL | CSTOPB;
  ttyios.c_cflag &= ~( PARENB );
  if(tcsetattr(ttyfd, TCSANOW, &ttyios ) == -1){
    fprintf(stderr, "stty error\n" );
    return -1;
  }
  tcflush(ttyfd, TCIOFLUSH);

  SendCMD("\n");
  
  if(!strcmp(argv[1], "-ram"))
    {
      srcfp = fopen(argv[2], "r");
      if(!srcfp)
	{
	  printf("Opening user PGM failed\n");
	  return -1;
	} 
      
      usleep(100000);
      tcflush(ttyfd, TCIOFLUSH);
      usleep(100000);
      tcflush(ttyfd, TCIOFLUSH);
      printf("Writing...\n");
      if(ReadMOT(srcfp, bSend))
	{
	  printf("Failed\n");
	  return -1;
	}
      fclose(srcfp);
    }
  close(ttyfd);
  return 0;
}

unsigned short MakeCheckSum(unsigned char *buf, int cnt)
{
  unsigned short sum = 0;
  for(;cnt > 0; cnt--)
    {
      sum += *(buf++);
    }
  return sum;
}

int SendCMD(const char *fmt, ...)
{
  va_list ap;
  char buf[256];
  char c;
  int i;

  va_start(ap, fmt);
  vsnprintf(buf, 256, fmt, ap);
  for(i = 0; i < strlen(buf); i++)  {
	write(ttyfd, buf + i, 1);
    tcdrain(ttyfd);
  }
  printf("%s", buf);

  while(read(ttyfd, &c, 1), c != '\n') putchar(c);
  putchar(c);
  va_end(ap);
  return 0;
}

int BlkSend(char *buf, unsigned int start, int cnt)
{
  unsigned short sum;
  char c;

  if(start <= 0x400000) return 0;
  sum = MakeCheckSum(buf, cnt);
  SendCMD("load %X %X %X 0\n", (unsigned int)start, (unsigned int)cnt, (unsigned int)sum);
  write(ttyfd, ">", 1);
  usleep(1000);
  write(ttyfd, buf, cnt);
  usleep(100000);
  while(read(ttyfd, &c, 1), c != '\n') putchar(c);
  printf("\n");
  tcflush(ttyfd, TCIOFLUSH);
//  AwaitACK();
  return 0;
}

int bSend (unsigned char data, unsigned int address, int isflush)
{
  static unsigned char buf[SENDSIZE];
  static int cnt = 0;
  static unsigned int start = 0xffffff;

  if((start + cnt == address) && cnt && !isflush)
    {
      buf[cnt++] = data;
      if(cnt < SENDSIZE) return 0;
      BlkSend(buf, start, cnt);
      cnt = 0;
    }
  else
    {
      if(cnt) BlkSend(buf, start, cnt);
      cnt = 0;
      if(!isflush)
	{
	  buf[0] = data;
	  start = address;
	  cnt = 1;
	}
    }
  return 0;
}

int ReadMOT(FILE *fp, int (*bwrite)(unsigned char data, unsigned int address, int isflush))
{
  char cbuf[256];
  unsigned int length, address, byte;
  int n;
  int ret;

  while(fgets(cbuf, 256, fp))
    {
      if(cbuf[0] != 'S')
	{
	  printf("Bad S Format\n");
	  return -1;
	}
      switch(cbuf[1])
	{
	case '0':
	  break;
	case '1':
 	  if(sscanf(cbuf, "S1%2x%4x", &length, &address) != 2)
	    {
	      printf("Bad S Format\n");
	      return -1;
	    }
	  length -= 3;
	  for(n = 0; n < length; n++)
	    {
	      if(sscanf(cbuf + 2*n + 8, "%2x", &byte) != 1)
		{
		  printf("Bad S Format\n");
		  return -1;
		}
	      if((ret = bwrite(byte, address++, 0)))
		{
		  printf("Writing Failed\n");
		  return ret;
		}
	    }
	  break;
	case '2':
 	  if(sscanf(cbuf, "S2%2x%6x", &length, &address) != 2)
	    {
	      printf("Bad S Format\n");
	      return -1;
	    }
	  length -= 4;
	  for(n = 0; n < length; n++)
	    {
	      if(sscanf(cbuf + 2*n + 10, "%2x", &byte) != 1)
		{
		  printf("Bad S Format\n");
		  return -1;
		}
	      if((ret = bwrite(byte, address++, 0)))
		{
		  printf("Writing Failed\n");
		  return ret;
		}
	    }
	  break;
	case '3':
 	  if(sscanf(cbuf, "S3%2x%8x", &length, &address) != 2)
	    {
	      printf("Bad S Format\n");
	      return -1;
	    }
	  length -= 5;
	  for(n = 0; n < length; n++)
	    {
	      if(sscanf(cbuf + 2*n + 12, "%2x", &byte) != 1)
		{
		  printf("Bad S Format\n");
		  return -1;
		}
	      if((ret = bwrite(byte, address++, 0)))
		{
		  printf("Writing Failed\n");
		  return ret;
		}
	    }
	  break;
	case '7':
	case '8':
	case '9':
	  printf("Flushing buffer\n");
	  if((ret = bwrite(byte, 0, 1)))
	    {
	      printf("Writing Failed\n");
	      return ret;
	    }
          SendCMD("run 400100\n");
	  return 0;
	default:	
	  printf("Bad S Format\n");
	  return -1;
	}
    }
  printf("Bad S Format\n");
  return -1;
}

