#include "config.h"
#include "commands.h"
#include "support.h"
#include "io.h"
#include "sci.h"

unsigned long __attribute__ ((section(".intram"))) locaddr;

typedef void (*fp)(void);

#ifdef RAMSTART
int CMDCacheOn (const char *buf, int fd)
{
unsigned long *address;

  for(address = (unsigned long *)0xfffff000; 
  	address < (unsigned long *)0xfffff400; address++)
  {
  	*address = 0;
  }
  CCR.BIT.CECS1 = 1;
  
  fdprintf(fd, "Cache successfully enabled" NEWLINE);
  return 0;  
}
#endif

int CMDReset (const char *buf, int fd)
{
fp p;
  p = *((fp*)0x0);
  p();
  return 0;  
}

int CMDSetBaudRate (const char *buf, int fd)
{
unsigned long rate;
int stop;
int n;
  n = sscanf(buf, "%lu %d", &rate, &stop);
  rate = rate / 100;
  if(n < 1) return -1;
  if(n < 2) stop = 0;
  fdprintf(fd, "rate:%lu, stop:%d" NEWLINE, rate, stop);
  InitSCI1(0, 0, stop, rate);
  return 0;  
}

int CMDDump (const char *buf, int fd)
{
  unsigned long address, bytes, i, c;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &bytes);
  if(n == 0) address = locaddr;
  if(n < 2) bytes = 0x100;
  fdprintf(fd, "START: %lX, STOP: %lX, BYTES: %lX" NEWLINE, 
	   address, (unsigned long)(address + bytes - 1), bytes);
  for(i = address, c = 0; c < bytes; i++, c++)
    {
      if(c % 16 == 0)
	fdprintf(fd, "%lX :", address);
      fdprintf(fd, " %.2X", (unsigned int) *((unsigned char *) address++));
      if(c % 16 == 15)
	fdprintf(fd, NEWLINE);
    }
  fdprintf(fd, NEWLINE);
  return 0;
}

int CMDGoLoc (const char *buf, int fd)
{
  unsigned long address;
  int n;

  n = sscanf(buf, "%lx", &address);
  if(n != 1) return -1;
  locaddr = address;
  return 0;
}

int CMDWriteByte (const char *buf, int fd)
{
  unsigned long address, data;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &data);
  if(n != 2) return -1;
  *((unsigned char *) address) = (unsigned char) data;
  return 0;
}

int CMDWriteWord (const char *buf, int fd)
{
  unsigned long address, data;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &data);
  if(n != 2) return -1;
  *((unsigned short *) address) = (unsigned short) data;
  return 0;
}

int CMDSleep (const char *buf, int fd)
{
  unsigned int ms;
  int n;

  n = sscanf(buf, "%u", &ms);
  if(n != 1) return -1;
  sleepinms(ms);
  fdprintf(fd, "OK" NEWLINE);
  return 0;
}

int CMDLoad (const char *buf, int fd)
{
  int n;
  unsigned long address, bytes, sum, start;
  char c;

  start = 0;
  n = sscanf(buf, "%lx%lx%lx%lx", &address, &bytes, &sum, &start);
  if(n < 3) return -1;
  if(n < 4) start = 0;

  do {read(fd, &c, 1);} while(c != '>');

  if(read(fd, (char *) address, bytes) != (int)bytes)
    {
      fdprintf(fd, "COM ERROR" NEWLINE);
      return 0;
    }
  if(sum != makesum((void *)address, bytes))
    {
      fdprintf(fd, "CHECK SUM ERROR" NEWLINE);
      return 0;
    }
  if(start)
    {
      ((fp)address)(); 
    }
  fdprintf(fd, "CHECK SUM OK" NEWLINE);
  return 0;
}

int CMDRun (const char *buf, int fd)
{
  int n;
  unsigned long address;

  n = sscanf(buf, "%lx", &address);
  if(n != 1) return -1;
  ((fp)address)();
  return 0;
}

int CMDMemtest (const char *buf, int fd)
{
  int n;
  unsigned long start;
  unsigned long bytes;
  volatile unsigned long *addr;
  unsigned long old;
  
  n = sscanf(buf, "%lx%lx", &start, &bytes);
  if(n != 2) return -1;
  for(addr = (unsigned long *)start; addr < (unsigned long *)(start + bytes); addr++)
  {
  	old = *addr;
  	*addr = (unsigned long)addr;
  	if(*addr != (unsigned long)addr)
	{
		fdprintf(fd, "Memtest failed addr:%lx, src:%lx, ret:%lx"  NEWLINE,
		(unsigned long)addr, (unsigned long)addr, *addr);
		return 0;
	}
  	*addr = ~(unsigned long)addr;
  	if(*addr != ~(unsigned long)addr)
	{
		fdprintf(fd, "Memtest failed addr:%lx, src:%lx, ret:%lx"  NEWLINE,
		(unsigned long)addr, ~(unsigned long)addr, *addr);
		return 0;
	}
	*addr = old;
  }
  fdprintf(fd, "Memtest succeeded" NEWLINE);
  return 0;
}
