#include "config.h"
#include "commands.h"
#include "support.h"

#include "gpib.h"

//Read ring buffer
#define MAX_GPIB_RDBUF 2048
char GPIBRDBuf[MAX_GPIB_RDBUF];
int GPIBRDBufCnt;
int GPIBRDBufStart;
#define MAX_GPIB_WRBUF 2048
char GPIBWRBuf[MAX_GPIB_WRBUF];
int GPIBWRBufCnt;

// Hardware config.
#define DR_GPIB_3SOE PB.DR.BIT.B4
#define IOR_GPIB_3SOE PFC.PBIOR.BIT.B4

#define DR_GPIB_SRQ_OUT PA.DR.BIT.B19
#define IOR_GPIB_SRQ_OUT PFC.PAIORH.BIT.B19
#define DR_GPIB_NDAC_OUT PB.DR.BIT.B2
#define IOR_GPIB_NDAC_OUT PFC.PBIOR.BIT.B2
#define DR_GPIB_NRFD_OUT PB.DR.BIT.B3
#define IOR_GPIB_NRFD_OUT PFC.PBIOR.BIT.B3
#define DR_GPIB_DAV_OUT PB.DR.BIT.B5
#define IOR_GPIB_DAV_OUT PFC.PBIOR.BIT.B5
#define DR_GPIB_EOI_OUT PA.DR.BIT.B9
#define IOR_GPIB_EOI_OUT PFC.PAIORL.BIT.B9
#define DR_GPIB_DIO_OUT (PE.DR.BYTE.L)
#define IOR_GPIB_DIO_OUT PFC.PEIOR.BYTE.L
#define DR_GPIB_DIO_IN (PF.DR.BYTE)
#undef IOR_GPIB_DIO_IN
#define DR_GPIB_ATN_IN PA.DR.BIT.B6
#define IOR_GPIB_ATN_IN PFC.PAIORL.BIT.B6
#undef DR_GPIB_SRQ_IN
#define DR_GPIB_IFC_IN PE.DR.BIT.B15
#define IOR_GPIB_IFC_IN PFC.PEIOR.BIT.B15
#define DR_GPIB_NDAC_IN PE.DR.BIT.B13
#define IOR_GPIB_NDAC_IN PFC.PEIOR.BIT.B13
#define DR_GPIB_NRFD_IN PE.DR.BIT.B11
#define IOR_GPIB_NRFD_IN PFC.PEIOR.BIT.B11
#define DR_GPIB_DAV_IN PE.DR.BIT.B10
#define IOR_GPIB_DAV_IN PFC.PEIOR.BIT.B10
#define DR_GPIB_REN_IN PE.DR.BIT.B8
#define IOR_GPIB_REN_IN PFC.PEIOR.BIT.B8
#define DR_GPIB_EOI_IN PE.DR.BIT.B9
#define IOR_GPIB_EOI_IN PFC.PEIOR.BIT.B9

char Talker;
char Listener;
char PlaySerialPoll;
unsigned char PriAddr;

union {
	unsigned char BYTE;
	struct {
	unsigned char B7:1;
	unsigned char RQS:1;
	unsigned char ESB:1;
	unsigned char MAV:1;
	unsigned char B3:1;
	unsigned char B2:1;
	unsigned char B1:1;
	unsigned char B0:1;
	} BIT;
} SPB;

//software anti-noise
inline
unsigned char readLine(unsigned char line)
{
unsigned char tmp;
	do tmp = line; while(tmp != line);
	return tmp;
}

int initGPIB(unsigned char priaddr)
{
	PriAddr = priaddr;
	
	IOR_GPIB_REN_IN = 0;	
	IOR_GPIB_EOI_IN = 0;	
	IOR_GPIB_DAV_IN = 0;	
	IOR_GPIB_NRFD_IN = 0;	
	IOR_GPIB_NDAC_IN = 0;	
	IOR_GPIB_IFC_IN = 0;	
//	IOR_GPIB_SRQ_IN = 0;	
	IOR_GPIB_ATN_IN = 0;	
	if((DR_GPIB_NRFD_IN == 1) && (DR_GPIB_NDAC_IN == 1))
		return -1; // interface is not fitted
	
	closeGPIB();
	
	Talker = -1;
	Listener = -1;
	PlaySerialPoll = 0;
	GPIBRDBufCnt = 0;
	GPIBRDBufStart = 0;
	GPIBWRBufCnt = 0;
	
	while(readLine(DR_GPIB_IFC_IN) == 0);
	DR_GPIB_NRFD_OUT = 0;
	DR_GPIB_NDAC_OUT = 0;
	return open(DEV_GPIB, 0);
}

int closeGPIB(void) {
	DR_GPIB_NRFD_OUT = 1;
	IOR_GPIB_NRFD_OUT = 1;
	DR_GPIB_NDAC_OUT = 1;
	IOR_GPIB_NDAC_OUT = 1;
	DR_GPIB_DAV_OUT = 1;
	IOR_GPIB_DAV_OUT = 1;
	DR_GPIB_EOI_OUT = 1;
	IOR_GPIB_EOI_OUT = 1;
	DR_GPIB_SRQ_OUT = 1;
	IOR_GPIB_SRQ_OUT = 1;
		
	DR_GPIB_3SOE = 1;
	IOR_GPIB_3SOE = 1;
	return 0;
}

int writeGPIB (int cnt, const unsigned char *bytes) {
int i;
	for(i = 0; i < cnt; i++) {
		GPIBWRBuf[GPIBWRBufCnt + i] = bytes[i];
	}
	GPIBWRBufCnt += cnt;
	return cnt;
}
int readGPIB (int cnt, unsigned char *bytes) {
int i;
	for(i = 0; i < cnt; i++) {
		if(GPIBRDBufCnt == 0) break;
		bytes[i] = GPIBRDBuf[GPIBRDBufStart++];
		if(GPIBRDBufStart == MAX_GPIB_RDBUF) GPIBRDBufStart = 0;
		GPIBRDBufCnt--;
	}
	return i;
}

//acceptor handshake
int doAH(unsigned char *byte, char *atn, char *eoi)
{
	DR_GPIB_NRFD_OUT = 1; //ready for data
	while(readLine(DR_GPIB_DAV_IN) == 1);
	DR_GPIB_NRFD_OUT = 0;
	*byte = readLine(DR_GPIB_DIO_IN) ^ 0xff;
	if(atn) *atn = (readLine(DR_GPIB_ATN_IN) == 0);
	if(eoi) *eoi = (readLine(DR_GPIB_EOI_IN) == 0);
	DR_GPIB_NDAC_OUT = 1;
	while(readLine(DR_GPIB_DAV_IN) == 0);
	DR_GPIB_NDAC_OUT = 0;
	return 0;
}

//source handshake
int doSH(unsigned char byte, char eoi)
{
	DR_GPIB_NRFD_OUT = 1;
	DR_GPIB_DAV_OUT = 1;
	DR_GPIB_EOI_OUT = eoi ? 0 : 1;
	DR_GPIB_DIO_OUT = byte ^ 0xff;
	DR_GPIB_3SOE = 0;
	sleepinus(GPIB_T1);
	while(readLine(DR_GPIB_NRFD_IN) == 0); //wait until RFD
	if(readLine(DR_GPIB_ATN_IN) == 0) {
		DR_GPIB_3SOE = 1;
		DR_GPIB_NRFD_OUT = 0;
		return -1; //run away if atn is asserted
	}
	DR_GPIB_DAV_OUT = 0;
	if(readLine(DR_GPIB_ATN_IN) == 0) {
		DR_GPIB_3SOE = 1;
		DR_GPIB_NRFD_OUT = 0;
		return -1; //run away if atn is asserted
	}
	DR_GPIB_NRFD_OUT = 0;
	DR_GPIB_NDAC_OUT = 1;
	while(readLine(DR_GPIB_NDAC_IN) == 0); //wait until DAC
	DR_GPIB_DAV_OUT = 1;
	while(readLine(DR_GPIB_DAV_IN) == 0); //wait until DAV gets stable
	DR_GPIB_3SOE = 1;
	DR_GPIB_NDAC_OUT = 0;
	return 0;
}

void doGPIB(int fd)
{
unsigned char byte;
char atn, eoi;
int wridx, i;
  while(readLine(DR_GPIB_ATN_IN) == 0);
  while(readLine(DR_GPIB_DAV_IN) == 0);
  for(;;)
  {
	if(readLine(DR_GPIB_IFC_IN) == 0) {
		fd =initGPIB(PriAddr);
		if(fd < 0) return;
	}
	
	doAH(&byte, &atn, &eoi);
	
	if(atn)
	{
	  switch(byte)
	  {
	   case 0x18: //SPE
	   	PlaySerialPoll = 1;
		break;
	   case 0x19: //SPD
	   	PlaySerialPoll = 0;
		break;
	   case 0x3f: //UNL
	   	Listener = -1;
		break;
	   case 0x5f: //UNT
	   	Talker = -1;
		break;
	   case 0x04: //SDL
	   	if(Listener != PriAddr) break;
	   case 0x14: //DCL
		break;
	  }
	  if((byte >= 0x40) && (byte < 0x60)) {
	  	Talker = byte - 0x40;
		wridx = 0;
	  }
	  if((byte >= 0x20) && (byte < 0x40)) {
		Listener = byte - 0x20;
		wridx = 0;
	  }
	}
	if((Listener == PriAddr) && (Talker > 0))
	{
	  if(GPIBRDBufCnt >= MAX_GPIB_RDBUF)
	  	Listener = -1;
	  else
	  {
	  	i = GPIBRDBufCnt + GPIBRDBufStart;
		if(i >= MAX_GPIB_RDBUF) i -= MAX_GPIB_RDBUF;
		GPIBRDBuf[i] = byte;
		GPIBRDBufCnt++;
	  	if(eoi) doConsole(fd);
		 continue;
	   }
	}
	if((Talker == PriAddr) && (Listener > 0))
	{
		if(PlaySerialPoll)
		{
			SPB.BYTE = 0;
			SPB.BIT.MAV = GPIBWRBufCnt ? 1 : 0;
			if(doSH(SPB.BYTE, 0)) continue;
			Talker = -1;
			continue;
		}
		else
		{
		   for(; wridx < GPIBWRBufCnt; wridx++)
		   {
			if(doSH(GPIBWRBuf[wridx], (wridx==GPIBWRBufCnt - 1))) break;
		   }
		   if(wridx < GPIBWRBufCnt) continue;
		}
	}
	/*
	//do nothing
	DR_GPIB_NDAC_OUT = 1;
	IOR_GPIB_NRFD_OUT = 1;
	while((readLine(DR_GPIB_ATN_IN) == 1) && 
		(readLine(DR_GPIB_IFC_IN) == 1));
	DR_GPIB_NRFD_OUT = 0;
	IOR_GPIB_NRFD_OUT = 0;
	*/
    }
}

