#include <string.h>
#include "sci.h"
#include "gpib.h"
#include "config.h"
#include "io.h"

#ifndef NOUSE_OPEN
char __attribute__ ((section(".intram"))) is_sci1_opened = 0;

int open (const char *name, int flags)
{
    if(!strcmp(name, DEV_SCI1))
    {
      is_sci1_opened = 1;
      return FD_SCI1;
    }
    if(!strcmp(name, DEV_GPIB))
    {
      return FD_GPIB;
    }
  return -1;
}

int close (int fd)
{
  switch(fd)
  {
  case FD_SCI1:
      if(!is_sci1_opened)
	{
	  //	  errno = EBADF;
	  return -1;
	}

      CloseSCI1();
      is_sci1_opened = 0;
      return 0;
  case FD_GPIB:
      closeGPIB();
      return 0;
  }
  //  errno = EBADF;
  return -1;
}
#endif

int write (int fd, const char *buf, int count)
{
  int n;

  switch(fd)
  {
  case FD_SCI1:
#ifndef NOUSE_OPEN
      if(!is_sci1_opened)
	{
	  //	  errno = EBADF;
	  return -1;
	}
#endif
#ifdef SIM_DEBUG
      n = fwrite(buf, 1, count, stdout);
#else
      n = WriteSCI1(count, buf);
#endif
      return n;
  case FD_GPIB:
      n = writeGPIB(count, buf);
      return n;
    }
  //  errno = EBADF;
  return -1;
}

int read (int fd, char *buf, int count)
{
  int n;

  switch(fd)
  {
  case FD_SCI1:
#ifndef NOUSE_OPEN
      if(!is_sci1_opened)
	{
	  //	  errno = EBADF;
	  return -1;
	}
#endif
#ifdef SIM_DEBUG
      n = fread(buf, 1, count, stdin);
#else
      n = ReadSCI1(count, buf);
#endif
      return n;
  case FD_GPIB:
      n = readGPIB(count, buf);
      return n;
    }
  //  errno = EBADF;
  return -1;
}

#ifndef NOUSE_OPEN
int ioctl(int fd, int request)
{
  if(fd == FD_SCI1)
    {
      if(request == REQ_FLUSHRDRF)
	{
	  return FlushRDRFSCI1();
	}
    }
  return -1;
}
#endif
