#include "config.h"
#include "commands.h"
#include "support.h"
#include "io.h"
#include "sci.h"

/*
 * All data are transferred using DMA.
 * Each DMA transfer is triggered by timer (MTU).
 * As of ver. 1.1, DMA uses "tandem" start.
 * MTU4 triggers a dummy transferring of DMAC2 (CS2 area, ext. wait enabled).
 * MTU4 also negates ~WAIT terminal. This synchronizes DMAC2.
 * DMAC0 is also triggered, but the priority is lower than that of DMAC2.
 * Thus, DMAC0 is started just after DMAC2 finishes dummy transferring.
 * DMAC0 is 2-state access.
 * MTU4 is free-running counter.
 */
 
#define DMA_STANBY_TIME 10
#define DMA_DUMMY_CYCLES 1

//  DM = 0 dst. fixed  
//  SM = 0 src. fixed.  
//  TM = 1 burst  
//  RS = 0xA MTU4TGIA
//  TS = 1 16bit
//  IE = 1
#define DMAC2_ACTIVE_CHCR_L 0x0A2CuL

typedef union {
	unsigned long lword;
	struct {
	  unsigned char byte4, byte3, byte2, byte1;
	} bytes;
	struct {
	  unsigned short msw, lsw;
	} words;
	struct {
	  unsigned char command;
	  unsigned char byte;
	  unsigned short word;
	} control;
} tPattern; //long word

#define PATTERNS_PER_BANK 100000
unsigned char PatternZipped[PATTERNS_ZIPPED_MAX];
tPattern Patterns[NUM_BANK][PATTERNS_PER_BANK];

tPattern *__attribute__ ((section(".intram"))) CurrentControl;
int __attribute__ ((section(".intram"))) CurrentBank;
#define MAX_LOOPS 20
tPattern * __attribute__ ((section(".intram"))) ControlAtDo[MAX_LOOPS];
int __attribute__ ((section(".intram"))) LoopCounters[MAX_LOOPS];
int __attribute__ ((section(".intram"))) LoopLevel;
int __attribute__ ((section(".intram"))) ErrByte;
char __attribute__ ((section(".intram"))) QAMOffset1, QAMOffset2;
short __attribute__ ((section(".intram"))) QAMLevel1, QAMLevel2;
char __attribute__ ((section(".intram"))) QAMDelay1, QAMDelay2;
unsigned long __attribute__ ((section(".intram"))) TimerU;

int UnzipPattern(unsigned char *src, int dstbank);
void pulseOff(void);

void InitPulser(void)
{
//PFCs, DRs
  PFC.PACRH.BIT.PA17MD = 1; //WAIT
  PFC.PACRL2.BIT.PA7MD = 2; //CS3
//  PFC.PACRL2.BIT.PA6MD = 2; //CS2
  
  DR_PULSE_OE = 1;
  IOR_PULSE_OE = 1; //out  
      
  BSC.BCR1.BIT.IOE = 0; //no multiplex IO
  BSC.BCR1.BIT.A3LG = 1; //CS3 long word access
//  BSC.BCR1.BIT.A3SZ = 0; //CS3 byte access
  BSC.BCR2.BIT.IW3 = 0; //CS3 no idle
  BSC.BCR2.BIT.CW3 = 0; //CS3 no idle
  BSC.BCR2.BIT.SW3 = 0; //CS3 no CS assert
  BSC.WCR1.BIT.W3 = 0; //CS3 no wait

  BSC.BCR1.BIT.A2LG = 1; //CS2 long word access
//  BSC.BCR1.BIT.A2SZ = 0; //CS2 byte access
  BSC.BCR2.BIT.IW2 = 0; //CS2 no idle
  BSC.BCR2.BIT.CW2 = 0; //CS2 no idle
  BSC.BCR2.BIT.SW2 = 0; //CS2 no CS assert
  BSC.WCR1.BIT.W2 = 1; //CS2 1 wait, ext wait enabled
  
  BSC.WCR2.BIT.DSW = 0; //CS area DMA ext. nowait
    
  MTU4.TCR.BIT.CKEG = 0;
  MTU4.TCR.BIT.TPSC = 0; //phi/1
  MTU4.TCR.BIT.CCLR = 0; //free-running.
  MTU4.TMDR.BIT.BFB = 0; //TGRB nonbuffered
  MTU4.TMDR.BIT.BFA = 0; //TGRA nonbuffered
  MTU4.TMDR.BIT.MD = 2; //PWM mode1
  MTU.TOER.BIT.OE4A = 1; //output enabled
  MTU4.TIOR.BIT.IOB = 6; //init. 1, to 1
  MTU4.TIOR.BIT.IOA = 5; //init. 1, to 0

  PE.DR.BIT.B12 = 1;
  PFC.PEIOR.BIT.B12 = 1; //out  
  PFC.PECR1.BIT.PE12MD = 1; //TIOC4A

  PFC.PACRL2.BIT.PA2MD = 2; //DREQ0
  PFC.PECR1.BIT.PE14MD = 2; //DACK0
  
  DMAC0.CHCR.WORD.L &= ~1u; // set DE = 0, first of all
  DMAC2.CHCR.WORD.L &= ~1u;
  
  DMAC.DMAOR.WORD; 
  DMAC.DMAOR.WORD = 0x0201; // priority, CH2>0>1>3    
  DMAC2.SAR = 0x0;
  DMAC2.DAR = (void *)0x800000uL; //CS2 area

  DMAC2.CHCR.LONG = 0;
//  .CHCR.BIT.AL = 1; //DACK low active  
//  .CHCR.BIT.SM = 1; //Src. incr.  
//  .CHCR.BIT.TM = 1; //burst  
//  .CHCR.BIT.RS = 2; //Single Address, to device
//  .CHCR.BIT.DS = 1; //DREQ neg. edge
//  .CHCR.BIT.TS = 2; //32bit
//  .CHCR.BIT.IE = 0;
  DMAC0.CHCR.LONG = 0x11270uL;
  
  INTC.IPRF.BIT._MTU4G = 1;
  INTC.IPRC.BIT._DMAC2 = 1;

  STI();
}

int CMDWriteAUX (const char *buf, int fd)
{
unsigned int data;
int port, n;
  n = sscanf(buf, "%d %x", &port, &data);
  if(n < 2) return -1;
  
  AUXOUT(port) = (unsigned char)data;
  DR_PULSE_OE = 0;
  
  return 0;
}

void pulseOff(void)
{
  CLI();
  DMAC0.CHCR.WORD.L &= ~1uL;
  DMAC2.CHCR.WORD.L &= ~1uL;
  MTU.TSTR.BIT.CST4 = 0;
  DR_PULSE_OE = 1;
}

int CMDPulseOff (const char *buf, int fd)
{
  fdprintf(fd, "Pulse ");
  pulseOff();
  fdprintf(fd, "Off" NEWLINE);
  return 0;
}

int CMDPulseLoad (const char *buf, int fd)
{
  int n;
  unsigned long bytes, sum;
  char c;

  n = sscanf(buf, "%lx", &bytes);
  if(n < 1) return -1;
  fdprintf(fd, "%lx" NEWLINE, bytes);

  read(fd, &c, 1);
  if(c != '>') return -1;

  if(bytes > PATTERNS_ZIPPED_MAX)
  {
      fdprintf(fd, "BUFFER OVERFLOW" NEWLINE);
      return 0;
  }
  if(read(fd, (char *)PatternZipped, bytes) != (int)bytes)
    {
      fdprintf(fd, "COM ERROR" NEWLINE);
      return 0;
    }
  if((n = UnzipPattern(PatternZipped, 0) ) ) {
  	fdprintf(fd, "Zip Parse Error line: %x, ret: %d" NEWLINE, ErrByte - (int)PatternZipped, n);
  	return -1;
  }
  sum = makesum((void *)PatternZipped, bytes);
  fdprintf(fd, "%lx" NEWLINE, sum);
  return 0;
}

int CMDPulseTestLoad (const char *buf, int fd)
{
int n;
unsigned char *zip = PatternZipped;
  *(zip++) = PATTERN_ZIPPED_COMMAND_PULSEON;  
  *(zip++) = PATTERN_ZIPPED_COMMAND_DO;  
  *(zip++) = 0;  
  *(zip++) = 10;  
  *(zip++) = PATTERN_ZIPPED_COMMAND_DO;  
  *(zip++) = 0;  
  *(zip++) = 3;  
  *(zip++) = PATTERN_ZIPPED_COMMAND_WAIT_LONG;  
  *(zip++) = 0x83u;
  *(zip++) = 0x00;  
  *(zip++) = 0; 
  *(zip++) = 25*4;
  *(zip++) = PATTERN_ZIPPED_COMMAND_DMA_SET;
  *(zip++) = 0xffu;  
  *(zip++) = 0xffu;  
  *(zip++) = PATTERN_ZIPPED_COMMAND_DMA_LSET_START + 60;
  *(zip++) = 0xeeu;
  *(zip++) = 0xeeu;
  *(zip++) = PATTERN_ZIPPED_COMMAND_DMA_LSET_START + 20;
  *(zip++) = 0xffu;
  *(zip++) = 0xffu;
  *(zip++) = PATTERN_ZIPPED_COMMAND_DMA_LSET_LONG;
  *(zip++) = 10;
  *(zip++) = 80;
  *(zip++) = 0x0;
  *(zip++) = 0x0;
  *(zip++) = PATTERN_ZIPPED_COMMAND_DMA_END;
  *(zip++) = PATTERN_ZIPPED_COMMAND_LOOP;
  *(zip++) = PATTERN_ZIPPED_COMMAND_WAIT_LONG;  
  *(zip++) = 0x83u;  
  *(zip++) = 0x00; 
  *(zip++) = 0; 
  *(zip++) = 200u; 
  *(zip++) = PATTERN_ZIPPED_COMMAND_LOOP_INF;
  *(zip++) = PATTERN_ZIPPED_COMMAND_END;

  if((n = UnzipPattern(PatternZipped, 0) ) ) {
  	fdprintf(fd, "Zip Parse Error line: %x, ret: %d" NEWLINE, ErrByte - (int)PatternZipped, n);
  	return -1;
  }
  
  fdprintf(fd, "Unzipping done." NEWLINE);
  
  return 0;
}


int CMDPulseOn (const char *buf, int fd)
{
  fdprintf(fd, "Pul");
  
  InitPulser();
  
  CurrentBank = 0;
  LoopLevel = -1;
  CurrentControl = Patterns[CurrentBank];
  QAMOffset1 = QAMOffset2 = 0;
  QAMLevel1 = QAMLevel2 = 0x100;
  QAMDelay1 = QAMDelay2 = 0;
  
  AUXOUT(0) = 0x0;
  AUXOUT(1) = 0xffu;
  AUXOUT(2) = 0x0;
  
  MTU.TSTR.BIT.CST4 = 0;
  MTU4.TGRA = 0x0;
  MTU4.TGRB = DMA_STANBY_TIME;
  MTU4.TCNT = 0xf000uL;
  
  MTU4.TSR.BIT.TGFA;
  MTU4.TSR.BIT.TGFA = 0;
  MTU4.TIER.BIT.TGIEA = 1;
  DMAC2.DMATCR = DMA_DUMMY_CYCLES;
  DMAC2.CHCR.WORD.L = DMAC2_ACTIVE_CHCR_L | 1uL;
  TimerU = 0;
  
  asm volatile ("" ::: "memory");
  
  fdprintf(fd, "se");
  
  //timer running start
  MTU.TSTR.BIT.CST4 = 1;
  
//  sleepinus(1);
//  processControls();
  
  fdprintf(fd, " On" NEWLINE);
  return 0;
}

inline int WriteAUX(unsigned char port, const unsigned char *buf, int len)
{
int i;
   if(len < 10) {
   	for(i = 0; i < len; i++)
		AUXOUT(port) = buf[i];
   }
   else {
	DMAC3.CHCR.WORD.L &= ~1u;
	DMAC3.SAR = (void *)buf;
	DMAC3.DMATCR = len;
	DMAC3.DAR = (void *)&AUXOUT(port);
//  DM = 0 dst. fixed  
//  SM = 1 src. incr.  
//  TM = 1 burst  
//  RS = 4 autorequest
//  TS = 0 8bit
//  IE = 0
	DMAC3.CHCR.LONG = 0x1420uL;
	DMAC3.CHCR.WORD.L |= 1u;
  }
  return 0;
}


int UnzipDMAPattern(unsigned char **src, tPattern *dststart, unsigned short startpattern)
{
tPattern *dsth1 = dststart; //pointer to DMA source of QAM1
tPattern *dsth2 = dststart; //pointer to DMA source of QAM2
tPattern *dstl = dststart; //pointer to DMA source of LSW
unsigned char *dasrc;
unsigned char x;
int size;
int i;
unsigned char rot;
unsigned char inv1, inv2;
char a, b;
unsigned char cmd, byte;
unsigned short LastPattern = startpattern;
  for(;;) {
  	cmd = *((*src)++);
  	switch(cmd) {
	case PATTERN_ZIPPED_COMMAND_DMA_END:
		for( i = 0; i < 2; i++) {
			(dstl++)->words.lsw = LastPattern;
		}
        x = 0x80u + QAMOffset1;
		for(; dsth1 < dstl; dsth1++) {
			dsth1->bytes.byte3 = x;
		}
        x = 0x80u + QAMOffset2;
		for(; dsth2 < dstl; dsth2++) {
			dsth2->bytes.byte4 = x;
		}
		for(; dstl < dsth1; dstl++) {
			dstl->words.lsw = LastPattern;
		}
		for(; dstl < dsth2; dstl++) {
			dstl->words.lsw = LastPattern;
		}
		size = dstl - dststart;
		return size;
	case PATTERN_ZIPPED_COMMAND_DMA_COPY_HBURST:
		dasrc = *src - 1;
		byte = *((*src)++); //phase
		switch(byte) {
		case 0:
			inv1 = 0; inv2 = 0; rot = 0; break;
		case 1:
			inv1 = 0xffu; inv2 = 0; rot = 1; break;
		case 2:
			inv1 = 0xffu; inv2 = 0xffu; rot = 0; break;
		case 3:
			inv1 = 0; inv2 = 0xffu; rot = 1; break;
		default:
			ErrByte = (int)(*src); return -4;
		}
		size = *((*src)++) * 0x100u;
		size += *((*src)++);
        x = 0x80u + QAMOffset1;
		for(; dsth1 < dststart + size + QAMDelay1; dsth1++) {
			dsth1->bytes.byte3 = x;
		}
        x = 0x80u + QAMOffset2;
		for(; dsth2 < dststart + size + QAMDelay2; dsth2++) {
			dsth2->bytes.byte4 = x;
		}
		dasrc -= *((*src)++) * 0x100u;
		dasrc -= *((*src)++);
		if(*dasrc != PATTERN_ZIPPED_COMMAND_DMA_HBURST) {ErrByte = (int)(*src); return -3;}
		dasrc++;
		size = *(dasrc++) * 0x100u;
		size += *(dasrc++);
        if(rot) {
            for( i = 1; i < size; i++) {
                a = (char)*(++dasrc);
                b = (char)*(++dasrc);
                x = ((unsigned short)((short)b * QAMLevel1) + (unsigned short)0x7f80u)/0x100u;
                (dsth1++)->bytes.byte3 = (x^inv1) + QAMOffset1;
                x = ((unsigned short)((short)a * QAMLevel2) + (unsigned short)0x7f80u)/0x100u;
                (dsth2++)->bytes.byte4 = (x^inv2) + QAMOffset2;
            }
        }
        else {
            for( i = 1; i < size; i++) {
                a = (char)*(++dasrc);
                b = (char)*(++dasrc);
                x = ((unsigned short)((short)a * QAMLevel1) + (unsigned short)0x7f80u)/0x100u;
                (dsth1++)->bytes.byte3 = (x^inv1) + QAMOffset1;
                x = ((unsigned short)((short)b * QAMLevel2) + (unsigned short)0x7f80u)/0x100u;
                (dsth2++)->bytes.byte4 = (x^inv2) + QAMOffset2;
            }
        }
		break;
	case PATTERN_ZIPPED_COMMAND_DMA_LSET_LONG:
		size = *((*src)++) * 0x100u;
		size += *((*src)++);
		for( i = 0; i < size; i++) {
			(dstl++)->words.lsw = LastPattern;
		}
		LastPattern = *((*src)++) * 0x100u;
		LastPattern += *((*src)++);
		break;
	default:
		if( (cmd < PATTERN_ZIPPED_COMMAND_DMA_LSET_START) ||
			(cmd > PATTERN_ZIPPED_COMMAND_DMA_LSET_END) )
		{ErrByte = (int)(*src); return -1;}
		size = cmd - PATTERN_ZIPPED_COMMAND_DMA_LSET_START;
		for( i = 0; i < size; i++) {
			(dstl++)->words.lsw = LastPattern;
		}
		LastPattern = *((*src)++) * 0x100u;
		LastPattern += *((*src)++);
		break;		
	}
  }
  {ErrByte = (int)(*src); return -1;}
}

int DAC7614Pattern(unsigned char *dst, unsigned char addr, unsigned short x)
{
union {
  struct {
	char loaddac1:1;
	char cs1:1;
	char clk1:1;
	char sdi1:1;
	char loaddac2:1;
	char cs2:1;
	char clk2:1;
	char sdi2:1;
  } bits;
  unsigned char byte;
 } aux;
 
int i;
int cnt = 0;
	x |= (3 - (addr & 0x3)) * 0x4000u;
	aux.byte = 0xffu;
	if(addr >= 4)
		aux.bits.cs2 = 0;
	else
		aux.bits.cs1 = 0;
	for(i = 0; i < 16; i++) {
		aux.bits.sdi1 = (x & 0x8000u) ? 1 : 0;
		aux.bits.sdi2 = (x & 0x8000u) ? 1 : 0;
		aux.bits.clk1 = 0;
		aux.bits.clk2 = 0;
		*(dst++) = aux.byte;
		cnt++;
		aux.bits.clk1 = 1;
		aux.bits.clk2 = 1;
		*(dst++) = aux.byte;
		cnt++;
		x *= 2;
	}
	aux.bits.cs1 = 1;	
	aux.bits.cs2 = 1;
	aux.bits.loaddac1 = 0;
	aux.bits.loaddac2 = 0;
	*(dst++) = aux.byte;
	cnt++;
	return cnt;
}

int UnzipPattern(unsigned char *src, int dstbank)
{
int j;
unsigned char byte, cmd;
unsigned short word;
tPattern *dst = Patterns[dstbank];
  for(;;) {
  	if(dst >= Patterns[dstbank] + PATTERNS_PER_BANK) {
		ErrByte = (int)src; return -2;
	}
  	cmd = *(src++);
  	switch(cmd) {
	case PATTERN_ZIPPED_COMMAND_END:
		(dst++)->control.command = PATTERN_COMMAND_END;
		return 0;
	case PATTERN_ZIPPED_COMMAND_DMA_SET:
		word = *(src++) * 0x100u;
		word += *(src++);
		j = UnzipDMAPattern(&src, dst + 1, word);
		if(j < 0) {return j;}
		if(j == 0) break;
		dst->control.command = PATTERN_COMMAND_DMA_SET;
		(dst++)->control.word = j;
		dst += j;
		break;
	case PATTERN_ZIPPED_COMMAND_DMA_HBURST:
		j = *(src++) * 0x100u;
		j += *(src++);
		src += j*2;
		break;	
	case PATTERN_ZIPPED_COMMAND_WAIT_LONG_LONG:
		(dst)->control.command = PATTERN_COMMAND_WAIT;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst++)->control.word = word;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst)->words.msw = word;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst++)->words.lsw = word;		
		break;
	case PATTERN_ZIPPED_COMMAND_WAIT_LONG:
		(dst)->control.command = PATTERN_COMMAND_WAIT;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst++)->control.word = word;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst++)->lword = word;
		break;
	case PATTERN_ZIPPED_COMMAND_WAIT:
		(dst)->control.command = PATTERN_COMMAND_WAIT;
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst++)->control.word = word;
		(dst++)->lword = 0;
		break;
	case PATTERN_ZIPPED_COMMAND_AUX1:
		byte = *(src++);
		(dst)->control.command = PATTERN_COMMAND_AUX;
		(dst)->control.byte = 0;
		(dst++)->control.word = 1;
		*((unsigned char *)dst) = byte;
		dst++;
		break;
	case PATTERN_ZIPPED_COMMAND_AUX3:
		byte = *(src++);
		(dst)->control.command = PATTERN_COMMAND_AUX;
		(dst)->control.byte = 2;
		(dst++)->control.word = 1;
		*((unsigned char *)dst) = byte;
		dst++;
		break;
	case PATTERN_ZIPPED_COMMAND_AUX2_DA:
		byte = *(src++);
		word = *(src++) * 0x100u;
		word += *(src++);
		byte = DAC7614Pattern((unsigned char*)(dst + 1), byte, word);
		(dst)->control.command = PATTERN_COMMAND_AUX;
		(dst)->control.byte = 1;
		(dst++)->control.word = byte;
		dst += (byte + 3) / 4;
		break;
	case PATTERN_ZIPPED_COMMAND_DO:
		word = *(src++) * 0x100u;
		word += *(src++);
		(dst)->control.command = PATTERN_COMMAND_DO;
		(dst++)->control.word = word; //counter
		break;
	case PATTERN_ZIPPED_COMMAND_LOOP:
		(dst++)->control.command = PATTERN_COMMAND_LOOP;
		break;
	case PATTERN_ZIPPED_COMMAND_LOOP_INF:
		(dst++)->control.command = PATTERN_COMMAND_LOOP_INF;
		break;
	case PATTERN_ZIPPED_COMMAND_BREAKPOINT:
		(dst++)->control.command = PATTERN_COMMAND_BREAKPOINT;
		break;
	case PATTERN_ZIPPED_COMMAND_PULSEON:
		(dst++)->control.command = PATTERN_COMMAND_PULSEON;
		break;
	case PATTERN_ZIPPED_COMMAND_SET_DA_TUNE_OFFSET:
		QAMOffset1 = (char)*(src++);
		QAMOffset2 = (char)*(src++);
		break;
	case PATTERN_ZIPPED_COMMAND_SET_DA_TUNE_LEVEL:
		QAMLevel1 = 0x100 + (char)*(src++);
		QAMLevel2 = 0x100 + (char)*(src++);
		break;
	case PATTERN_ZIPPED_COMMAND_SET_DA_TUNE_DELAY:
		QAMDelay1 = (char)*(src++);
		QAMDelay2 = (char)*(src++);
		break;
	default:
		ErrByte = (int)src; return -1;
	}
  }
}

inline void processControls(void)
{
unsigned char byte, cmd;
unsigned short word;
int setwait = 0;
	for(;;)
	{
		cmd = CurrentControl->control.command;
		byte = CurrentControl->control.byte;
		word = CurrentControl->control.word;
		CurrentControl++;
		switch(cmd)
		{
		case PATTERN_COMMAND_DMA_SET:
            if((DMAC0.CHCR.WORD.L & 1) && !(DMAC0.CHCR.WORD.L & 2)) {
                pulseOff();
                fdprintf(FD_SCI1, "DMAC0ERR1" NEWLINE);
                return;
             }
            DMAC0.CHCR.WORD.L &= ~3u;
			if(!setwait) {
                pulseOff();
                fdprintf(FD_SCI1, "NOWAITERR1" NEWLINE);
                return;
             }
			DMAC0.SAR = (unsigned long*)CurrentControl;
    			DMAC0.DMATCR = word;
			DMAC0.CHCR.WORD.L |= 1u;
			CurrentControl += word;
      		return;
		case PATTERN_COMMAND_WAIT:
			if(setwait) {CurrentControl--; return;}
			TimerU = (CurrentControl++)->lword;
             TimerU *= 2;
             if((TimerU >= 1) && (word < 0x8000u)) {
                word += 0x8000u;
                TimerU--;
             }
             MTU4.TGRA += word;
			if(!TimerU) {
                if((unsigned short)MTU4.TGRA - (unsigned short)MTU4.TCNT < 10u*28u/4u) {
                    pulseOff();
                    fdprintf(FD_SCI1, "TOOSMALLWAITERR1" NEWLINE);
                    return;
                 }
				MTU4.TGRB = MTU4.TGRA + DMA_STANBY_TIME;
				DMAC2.CHCR.WORD.L &= ~3u;
				DMAC2.DMATCR = DMA_DUMMY_CYCLES;
                 DMAC2.CHCR.WORD.L = DMAC2_ACTIVE_CHCR_L | 1uL;
			}
			else {
                 // clear trigger source not to prevent MTU4TGIA.
                 DMAC2.CHCR.WORD.L = 0uL;
				MTU4.TGRB = MTU4.TGRA;
			}
			setwait = 1;
			break;
		case PATTERN_COMMAND_AUX:
			WriteAUX(byte, (unsigned char*)CurrentControl, word);
			CurrentControl += (word + 3) / 4;
			break;
		case PATTERN_COMMAND_DO:
			LoopLevel++;
			ControlAtDo[LoopLevel] = CurrentControl;
			LoopCounters[LoopLevel] = word;
			break;
		case PATTERN_COMMAND_LOOP:
			if(--LoopCounters[LoopLevel] == 0) {
				LoopLevel--;
				break;
			}
		case PATTERN_COMMAND_LOOP_INF:
			CurrentControl = ControlAtDo[LoopLevel];
			break;
		case PATTERN_COMMAND_BREAKPOINT:
			break;
		case PATTERN_COMMAND_PULSEON:
			DR_PULSE_OE = 0;
			break;
        default:
             pulseOff();
             fdprintf(FD_SCI1, "NOSUCHCMDERR1" NEWLINE);
             return;
		case PATTERN_COMMAND_END:
             pulseOff();
			return;
		}
	}
}

//#pragma interrupt
void DMAC2DEI(void)
{
  DMAC2.CHCR.WORD.L &= ~3u;
  MTU4.TSR.BIT.TGFA;
  MTU4.TSR.BIT.TGFA = 0;
  processControls();
  return;
}

//#pragma interrupt
void MTU4TGIA(void)
{
   MTU4.TSR.BIT.TGFA;
   MTU4.TSR.BIT.TGFA = 0;
   
   MTU4.TGRA += 0x8000uL;
   
   TimerU--;
   if(!TimerU)
   {
       unsigned short grab = MTU4.TGRA + DMA_STANBY_TIME;
       MTU4.TGRB = grab;
       //setup dummy transferring
       DMAC2.CHCR.WORD.L &= ~0x3;
       DMAC2.DMATCR = DMA_DUMMY_CYCLES;
       DMAC2.CHCR.WORD.L = DMAC2_ACTIVE_CHCR_L | 1uL;
   }
   else {
       MTU4.TGRB = MTU4.TGRA;
   }
  return;
}
