#include "config.h"
#include "support.h"
#include "sci.h"
#include <string.h>

#ifndef NOUSE_OPEN
int InitSCI1 (int parity, int odd, int stop, unsigned int baudratep100)
{
  int dck, cks;
  unsigned int n;
  
  dck = 1;
  for(cks = 0; cks < 4; cks++)
    {
      n = ((unsigned long)CLOCK_IN_Hz) / (50 * baudratep100 * 32 * dck);
      n = (n + 1) / 2 - 1;
      if(n < 256) break;
      dck *= 4;
    }

  SCI1.SCR.BIT.TE = 0;
  SCI1.SCR.BIT.RE = 0;

  //no SCK, no interrupts
  SCI1.SCR.BYTE = 0x00;
  //ensure
  SCI1.SCR.BYTE = 0x00;

  //asynchronous, 8bit, noparity, even
  SCI1.SMR.BYTE = cks;
  if(parity) SCI1.SMR.BIT._PE = 1;
  if(odd) SCI1.SMR.BIT.OE = 1;
  if(stop) SCI1.SMR.BIT.STOP = 1;

  SCI1.BRR = n;

  sleepinus(2000000/(baudratep100*100));

  SCI1.SCR.BIT.TE = 1;
  FlushRDRFSCI1();
  return 0;
}

int CloseSCI1 (void)
{
  SCI1.SCR.BYTE = 0x00;
  return 0;
}

int FlushRDRFSCI1 (void)
{
  SCI1.SSR.BIT.ORER;
  SCI1.SSR.BIT.ORER = 0;
  SCI1.SSR.BIT.PER;
  SCI1.SSR.BIT.PER = 0;
  SCI1.SSR.BIT.FER;
  SCI1.SSR.BIT.FER = 0;
  SCI1.SSR.BIT.RDRF;
  SCI1.SSR.BIT.RDRF = 0;
  SCI1.SCR.BIT.RE = 0;
  SCI1.SCR.BIT.RE = 1;
  return 0;
}
#endif
  
int WriteSCI1 (int cnt, const unsigned char *bytes)
{
  int n;
  for(n = 0; n < cnt; n++)
    {
      while(!SCI1.SSR.BIT.TDRE);
      
      SCI1.TDR = *bytes++;
      SCI1.SSR.BIT.TDRE = 0;
    }
  while(!SCI1.SSR.BIT.TEND);
      
  return cnt;
}

int ReadSCI1 (int cnt, unsigned char *bytes)
{
  int n;
  const unsigned char ssr_rflags = 0x78;
  const char msg_orer[] = "SCI1 ORER";
  const char msg_fer[] = "SCI1 FER";
  const char msg_per[] = "SCI1 PER";
  for(n = 0; n < cnt; n++)
    {
      while((SCI1.SSR.BYTE & ssr_rflags) == 0) ;

    	  if(SCI1.SSR.BIT.ORER)
         {
              WriteSCI1(sizeof(msg_orer), msg_orer);
              SCI1.SSR.BYTE &= ~ssr_rflags;
              return n;
         }
       if(SCI1.SSR.BIT.FER)
         {
              WriteSCI1(sizeof(msg_fer), msg_fer);
              SCI1.SSR.BYTE &= ~ssr_rflags;
              return n;
         }
       if(SCI1.SSR.BIT.PER)
    	    {
              WriteSCI1(sizeof(msg_per), msg_per);
              SCI1.SSR.BYTE &= ~ssr_rflags;
              return n;
    	    }
      *bytes++ = SCI1.RDR;
      SCI1.SSR.BIT.RDRF = 0;
    }
  return cnt;
}
